/*
 **********************************************************************
 *     irqmgr.c - IRQ manager for emu10k1 driver
 *     Copyright 1999, 2000 Creative Labs, Inc.
 *
 **********************************************************************
 *
 *     Date                 Author          Summary of changes
 *     ----                 ------          ------------------
 *     October 20, 1999     Bertrand Lee    base code release
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

#include "hwaccess.h"
#include "cardmi.h"
#include "cardmo.h"
#include "irqmgr.h"

/* Interrupt handler */
static void emu10k1_interrupt(int nIRQ, void *pvDevID, struct pt_regs *pRegs)
{
	struct sblive_irq *irq_ptr = (struct sblive_irq *) pvDevID;
	struct sblive_hw *sb_hw = irq_ptr->sb_hw;
	u32 irqstatus, ptr, tmp;

	if (sb_hw->hw_irq.ID != EMU10K1_INTERRUPT_ID)
		return;

	DPD(4, "emu10k1_interrupt called, nIRQ =  %u\n", nIRQ);

	/* Preserve PTR register */
	ptr = sblive_readfn0(sb_hw, PTR);

	/*
	 ** NOTE :
	 ** We do a 'while loop' here cos on certain machines, with both
	 ** playback and recording going on at the same time, IRQs will
	 ** stop coming in after a while. Checking IPND indeed shows that
	 ** there are interrupts pending but the PIC says no IRQs pending.
	 ** I suspect that some boards need edge-triggered IRQs but are not
	 ** getting that condition if we don't completely clear the IPND
	 ** (make sure no more interrupts are pending).
	 ** - Eric
	 */

	irqstatus = sblive_readfn0(sb_hw, IPR);

	do
	{
		DPD(4, "irq status %x\n", irqstatus);

		tmp = irqstatus;

		if (irqstatus & IRQTYPE_TIMER)
		{
			sblive_timerIrqCallback(sb_hw);
			irqstatus &= ~IRQTYPE_TIMER;
		}

		if (irqstatus & IRQTYPE_RECORD)
		{
			sblive_waveinIrqCallback(sb_hw);
			irqstatus &= ~IRQTYPE_RECORD;
		}

		if (irqstatus & IRQTYPE_MPUIN)
		{
			sblive_mpuinIrqCallback(sb_hw);
			irqstatus &= ~IRQTYPE_MPUIN;
		}

		if (irqstatus & IRQTYPE_MPUOUT)
		{
			sblive_mpuoutIrqCallback(sb_hw);
			irqstatus &= ~IRQTYPE_MPUOUT;
		}

		if(irqstatus)
			sblive_irqmgrDisableIrq(sb_hw, irqstatus);

		sblive_writefn0(sb_hw, IPR, tmp);

	} while ((irqstatus = sblive_readfn0(sb_hw, IPR)));

	sblive_writefn0(sb_hw, PTR, ptr);

	return;
}

/************************************************************************
 *
 *   int    sblive_irqmgrInit(struct sblive_hw * sb_hw)
 *
 *   ENTRY
 *       sb_hw -   Card object
 *
 *   RETURNS
 *       SUCCESS -   CTSTATUS_SUCCESS
 *       FAILURE -   CTSTATUS_ERROR
 *
 *   ABOUT
 *       Initialization
 *
 ************************************************************************/
int sblive_irqmgrInit(struct sblive_irq *irq_ptr)
{
	unsigned long flags;

	irq_ptr->status = 0;
	irq_ptr->ID = EMU10K1_INTERRUPT_ID;

	flags = SA_SHIRQ;

	/* Reserve IRQ Line */
	if (request_irq(irq_ptr->irq, emu10k1_interrupt, flags, "emu10k1", irq_ptr))
	{
		DPD(2, "emu10k1: irq %u in use\n", (unsigned int) irq_ptr->irq);
		irq_ptr->status &= ~FLAGS_ENABLED;
		DPF(2, "Error initializing IRQ\n");
		return CTSTATUS_ERROR;
	}

	irq_ptr->status |= (FLAGS_ENABLED | FLAGS_AVAILABLE);

	if (irq_ptr->status & (FLAGS_ENABLED | FLAGS_AVAILABLE))
		irq_ptr->status &= (~FLAGS_AVAILABLE);
	else
	{
		DPF(2, "IRQ acquire failure\n");
		return CTSTATUS_INUSE;
	}

	DPD(2, "IRQ : Acquired channel %u\n", irq_ptr->irq);
	return CTSTATUS_SUCCESS;
}


/************************************************************************
 *
 *   int    sblive_irqmgrExit(struct sblive_irq *irq_ptr)
 *
 *   ENTRY
 *       irq_ptr -   Card irq object
 *
 *   RETURNS
 *       Always returns CTSTATUS_SUCCESS
 *
 *   ABOUT
 *       Shutdown code
 *
 ************************************************************************/
int sblive_irqmgrExit(struct sblive_irq *irq_ptr)
{
	irq_ptr->status |= FLAGS_AVAILABLE;
	free_irq(irq_ptr->irq, irq_ptr);
	irq_ptr->status &= ~FLAGS_ENABLED;

	return CTSTATUS_SUCCESS;
}

/************************************************************************
 *
 *   int    sblive_irqmgrEnableIrq(struct sblive_hw * sb_hw, u16 wIrqType)
 *
 *   ENTRY
 *       sb_hw -   Card object
 *       wIrqType-   type of service to enable interrupt for
 *
 *   RETURNS
 *       Always returns CTSTATUS_SUCCESS
 *
 *   ABOUT
 *       Enables the specified irq service
 *
 ************************************************************************/
int sblive_irqmgrEnableIrq(struct sblive_hw *sb_hw, u32 irqtype)
{
	/*
	 * TODO :
	 * put protection here so that we don't accidentally
	 * screw-up another cardxxx objects irqs
	 */

	DPD(4, "sblive_irqmgrEnableIrq %x\n", irqtype);
	sblive_wrtmskfn0(sb_hw, INTE, irqtype, ENABLE);

	return CTSTATUS_SUCCESS;
}


/************************************************************************
 *
 *   int    sblive_irqmgrDisableIrq(struct sblive_hw * sb_hw, u16 wIrqType)
 *
 *   ENTRY
 *       sb_hw -   Card object
 *       wIrqType-   type of service to disable interrupt for
 *
 *   RETURNS
 *       Always returns CTSTATUS_SUCCESS
 *
 *   ABOUT
 *       Disables the specified irq service
 *
 ************************************************************************/

int sblive_irqmgrDisableIrq(struct sblive_hw *sb_hw, u32 irqtype)
{
	/*
	 * TODO :
	 * put protection here so that we don't accidentally
	 * screw-up another cardxxx objects irqs
	 */

	DPD(4, "sblive_irqmgrDisableIrq %x\n", irqtype);
	sblive_wrtmskfn0(sb_hw, INTE, irqtype, DISABLE);

	return CTSTATUS_SUCCESS;
}

