/*
 **********************************************************************
 *     midi.c - /dev/midi interface for emu10k1 driver
 *     Copyright 1999, 2000 Creative Labs, Inc.
 *
 **********************************************************************
 *
 *     Date                 Author          Summary of changes
 *     ----                 ------          ------------------
 *     October 20, 1999     Bertrand Lee    base code release
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

#ifdef MODULE
#define __NO_VERSION__
#include <linux/module.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include "hwaccess.h"
#include "mycommon.h"
#include "cardmo.h"
#include "cardmi.h"
#include "midi.h"


static spinlock_t midi_spinlock = SPIN_LOCK_UNLOCKED;

static int emu10k1_midi_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct sblive_hw *sb_hw = sblive_devs;
	struct sblive_mididevice *midi_dev;

	DPF(2, "emu10k1_midi_open() called\n");

	/* Check for correct device to open */
	while (sb_hw && sb_hw->midi_num != minor)
	  sb_hw = sb_hw->next;
	if (!sb_hw)
	  return -ENODEV;

	/* Wait for device to become free */
	down(&sb_hw->open_sem);
	while (sb_hw->open_mode & (file->f_mode << FMODE_MIDI_SHIFT)) 
	{
		if (file->f_flags & O_NONBLOCK) 
		{
			up(&sb_hw->open_sem);
			return -EBUSY;
		}
		
		up(&sb_hw->open_sem);
		interruptible_sleep_on(&sb_hw->open_wait);
		
		if (signal_pending(current))
		  return -ERESTARTSYS;
		
		down(&sb_hw->open_sem);
	}

	midi_dev = (struct sblive_mididevice *) kmalloc(sizeof(*midi_dev), GFP_KERNEL);
	if (midi_dev == NULL) 
	{
		DPF(2, " MIDIDEVICEOBJ alloc fail.");
		return -EINVAL;
	}
	
	DPD(3, "kmalloc: [%p]\n", midi_dev);
	
	midi_dev->sb_hw = sb_hw;
	midi_dev->mistate = MIDIIN_STATE_STOPPED;
	init_waitqueue_head(&midi_dev->oWait);
	init_waitqueue_head(&midi_dev->iWait);
	midi_dev->ird = 0;
	midi_dev->iwr = 0;
	midi_dev->icnt = 0;
	midi_dev->pmiHdrList = NULL;

	if (file->f_mode & FMODE_READ) 
	{
		struct midi_openinfo dsCardMidiOpenInfo;
		struct midi_hdr *midihdr1;
		struct midi_hdr *midihdr2;

		dsCardMidiOpenInfo.refdata = (unsigned long) midi_dev;

		if (sblive_mpuinOpen(sb_hw, &dsCardMidiOpenInfo) 
		    != CTSTATUS_SUCCESS) 
		{
			DPF(2, "sblive_mpuinOpen failed.\n");

			kfree(midi_dev);
			DPD(3, "kfree: [%p]\n", midi_dev);			

			return -ENODEV;
		}

		/* Add two buffers to receive sysex buffer */
		if (midiInAddBuffer(midi_dev, &midihdr1) != CTSTATUS_SUCCESS) 
		{
			kfree(midi_dev);
			DPD(3, "kfree: [%p]\n", midi_dev);

			return -ENODEV;
		}

		if (midiInAddBuffer(midi_dev, &midihdr2) != CTSTATUS_SUCCESS) 
		{
			osListRemove((struct sblive_list **) &midi_dev->pmiHdrList, (struct sblive_list *) midihdr1);

			kfree(midihdr1->lpData);
			DPD(3, "kfree: [%p]\n", midihdr1->lpData);

			kfree(midihdr1);
			kfree(midi_dev);

			DPD(3, "kfree: [%p]\n", midihdr1);
			DPD(3, "kfree: [%p]\n", midi_dev);

			return -ENODEV;
		}
	}

	if (file->f_mode & FMODE_WRITE) 
	{
		struct midi_openinfo dsCardMidiOpenInfo;

		dsCardMidiOpenInfo.refdata = (unsigned long) midi_dev;

		if (sblive_mpuoutOpen(sb_hw, &dsCardMidiOpenInfo)
		    != CTSTATUS_SUCCESS) 
		{
			DPF(2, "sblive_mpuoutOpen failed.\n");

			kfree(midi_dev);
			DPD(3, "kfree: [%p]\n", midi_dev);

			return -ENODEV;
		}
	}

	file->private_data = (void *) midi_dev;

	sb_hw->open_mode |= (file->f_mode << FMODE_MIDI_SHIFT) & (FMODE_MIDI_READ | FMODE_MIDI_WRITE);

	up(&sb_hw->open_sem);

	MOD_INC_USE_COUNT;
	return 0;
}

static int emu10k1_midi_release(struct inode *inode, struct file *file)
{
	struct sblive_mididevice *midi_dev = (struct sblive_mididevice *) file->private_data;
	struct sblive_hw *sb_hw = midi_dev->sb_hw;

	DPF(2, "emu10k1_midi_release() called\n");
	
	if (file->f_mode & FMODE_WRITE) 
	{
		if(!(file->f_flags & O_NONBLOCK)){

			while (!signal_pending(current) && (sb_hw->card_mpuout->firstmidiq !=NULL)){
				DPF(2, "Cannot close - buffers not empty\n");

				interruptible_sleep_on(&midi_dev->oWait);		

			}
		}

		sblive_mpuoutClose(sb_hw);
	}
	
	if (file->f_mode & FMODE_READ) 
	{
		struct midi_hdr *midihdr;

		if (midi_dev->mistate == MIDIIN_STATE_STARTED) 
		{
			sblive_mpuinStop(sb_hw);
			midi_dev->mistate = MIDIIN_STATE_STOPPED;
			DPF(2, "MIDI in stopped\n");
		}
		
		sblive_mpuinReset(sb_hw);
		sblive_mpuinClose(sb_hw);
		
		DPF(2, "MIDI in closed\n");
		
		while ((midihdr = (struct midi_hdr *) osListGetNext((struct sblive_list *) midi_dev->pmiHdrList, NULL))) 
		{
			osListRemove((struct sblive_list **) &midi_dev-> pmiHdrList, (struct sblive_list *) midihdr);

			kfree(midihdr->lpData);
			DPD(3, "kfree: [%p]\n", midihdr->lpData);

			kfree(midihdr);
			DPD(3, "kfree: [%p]\n", midihdr);
		}
	}
	
	kfree(midi_dev);

	DPD(3, "kfree: [%p]\n", midi_dev);

	down(&sb_hw->open_sem);
	sb_hw->open_mode &= ~((file->f_mode << FMODE_MIDI_SHIFT) & (FMODE_MIDI_READ | FMODE_MIDI_WRITE));
	up(&sb_hw->open_sem);
	wake_up(&sb_hw->open_wait);
	
	MOD_DEC_USE_COUNT;

	return 0;
}


static ssize_t emu10k1_midi_read(struct file *file, char *buffer, size_t count, loff_t *pos)
{
	struct sblive_mididevice *midi_dev = (struct sblive_mididevice *) file->private_data;
	ssize_t ret = 0;
	u16 cnt;
	unsigned long flags;

	DPD(2, "emu10k1_midi_read() called, count %d\n", count);

	if (pos != &file->f_pos)
	  return -ESPIPE;

	if (!access_ok(VERIFY_WRITE, buffer, count))
	  return -EFAULT;

	if (midi_dev->mistate == MIDIIN_STATE_STOPPED) 
	{
		if (sblive_mpuinStart(midi_dev->sb_hw) 
		    != CTSTATUS_SUCCESS) 
		{
			DPF(2, "sblive_mpuinStart failed.\n");
			return -EINVAL;
		}
		
		midi_dev->mistate = MIDIIN_STATE_STARTED;
	}

	while (count > 0) 
	{
		cnt = MIDIIN_BUFLEN - midi_dev->ird;
	
		spin_lock_irqsave(&midi_spinlock, flags);
	
		if (midi_dev->icnt < cnt)
		  cnt = midi_dev->icnt;
		
		spin_unlock_irqrestore(&midi_spinlock, flags);
		
		if (cnt > count)
		  cnt = count;
		
		if (cnt <= 0) 
		{
			if (file->f_flags & O_NONBLOCK)
			  return ret ? ret : -EAGAIN;
			DPF(2, " Go to sleep...\n");
			
			interruptible_sleep_on(&midi_dev->iWait);
			
			if (signal_pending(current))
				return ret ? ret : -ERESTARTSYS;
			
			continue;
		}
		
		if (copy_to_user(buffer, midi_dev->iBuf + midi_dev->ird, cnt)) {
			DPF(2, " copy_to_user failed.");
			return ret ? ret : -EFAULT;
		}
		
		midi_dev->ird += cnt;
		midi_dev->ird %= MIDIIN_BUFLEN;

		spin_lock_irqsave(&midi_spinlock, flags);

		midi_dev->icnt -= cnt;
		
		spin_unlock_irqrestore(&midi_spinlock, flags);
		
		count -= cnt;
		buffer += cnt;
		ret += cnt;

		if (midi_dev->icnt == 0)
			break;

	}
	
	return ret;
}


static ssize_t emu10k1_midi_write(struct file *file, const char *buffer, size_t count, loff_t *pos)
{
	struct sblive_mididevice *midi_dev = (struct sblive_mididevice *) file->private_data;
	struct midi_hdr *midihdr;
	ssize_t ret = 0;
	unsigned long flags;

	DPD(2, "emu10k1_midi_write() called, count=%d\n", count);

	if (pos != &file->f_pos)
	  return -ESPIPE;

	if (!access_ok(VERIFY_READ, buffer, count))
	  return -EFAULT;

	midihdr = (struct midi_hdr *) kmalloc(sizeof(struct midi_hdr *), GFP_KERNEL);
	if (midihdr == NULL) 
		return -EINVAL;

	DPD(3, "kmalloc: [%p]\n", midihdr);
	
	midihdr->bufferlength = count;
	midihdr->bytesrecorded = 0;
	midihdr->flags = 0;
	midihdr->next = NULL;
	
	midihdr->lpData = (u8 *) kmalloc(count, GFP_KERNEL);
	if (midihdr->lpData == NULL) 
	{
		DPF(2, "midihdr->lpData alloc failed\n");

		kfree(midihdr);
		DPD(3, "kfree: [%p]\n", midihdr);
	
		return -EINVAL;
	}

	DPD(3, "kmalloc: [%p]\n", midihdr->lpData);
	
	if (copy_from_user(midihdr->lpData, buffer, count)) 
	{
		kfree(midihdr->lpData);
		DPD(3, "kfree: [%p]\n", midihdr->lpData);

		kfree(midihdr);
		DPD(3, "kfree: [%p]\n", midihdr);
		
		return ret ? ret : -EFAULT;
	}

	spin_lock_irqsave(&midi_spinlock, flags);

	if (sblive_mpuoutAddBuffer(midi_dev->sb_hw, midihdr) != CTSTATUS_SUCCESS) 
	{
		DPF(2, "sblive_mpuoutAddBuffer failed.\n");

		kfree(midihdr->lpData);
		DPD(3, "kfree: [%p]\n", midihdr->lpData);

		kfree(midihdr);
		DPD(3, "kfree: [%p]\n", midihdr);
	
		spin_unlock_irqrestore(&midi_spinlock, flags);
		return -EINVAL;
	}

	spin_unlock_irqrestore(&midi_spinlock, flags);

	return count;
}


static unsigned int emu10k1_midi_poll(struct file *file, struct poll_table_struct *wait)
{
	DPF(2, "emu10k1_midi_poll() called\n");
	return 0;
}


int midiCallbackFn(unsigned long msg, unsigned long refdata, unsigned long *pmsg)
{
	struct sblive_mididevice *midi_dev = (struct sblive_mididevice *) refdata;
	struct midi_hdr *midihdr = NULL;
	unsigned long flags;
	int i;

	DPF(2, "midiCallbackFn!\n");
	
	spin_lock_irqsave(&midi_spinlock, flags);

	switch (msg) 
	{
	case ICARDMIDI_OUTLONGDATA:
		midihdr = (struct midi_hdr *) pmsg[2];

		kfree(midihdr->lpData);
		DPD(3, "kfree: [%p]\n", midihdr->lpData);

		kfree(midihdr);
		DPD(3, "kfree: [%p]\n", midihdr);

		wake_up(&midi_dev->oWait);

		break;

	case ICARDMIDI_INLONGDATA:
		DPF(2, " ICARDMIDI_INLONGDATA\n");
		midihdr = (struct midi_hdr *) pmsg[2];

		for (i = 0; i < midihdr->bytesrecorded; i++) {
			midi_dev->iBuf[midi_dev->iwr++] = midihdr->lpData[i];
			midi_dev->iwr %= MIDIIN_BUFLEN;
		}

		midi_dev->icnt += midihdr->bytesrecorded;
		
		if (midi_dev->mistate == MIDIIN_STATE_STARTED) 
		{
			initMidiHdr(midihdr);
			sblive_mpuinAddBuffer(midi_dev->sb_hw->card_mpuin, midihdr);
			wake_up(&midi_dev->iWait);
		}
		break;

	case ICARDMIDI_INDATA:
		DPF(2, " ICARDMIDI_INDATA\n");
		{
			u8 *pBuf = (u8 *) & pmsg[1];
			u16 bytesvalid = pmsg[2];

			for (i = 0; i < bytesvalid; i++) {
				midi_dev->iBuf[midi_dev->iwr++] = pBuf[i];
				midi_dev->iwr %= MIDIIN_BUFLEN;
			}

			midi_dev->icnt += bytesvalid;
		}
		
		wake_up(&midi_dev->iWait);
		break;

	default:		/* Unknown message */
		DPF(2, "Unknown message\n");
		return CTSTATUS_ERROR;
	}

	spin_unlock_irqrestore(&midi_spinlock, flags);
	
	return CTSTATUS_SUCCESS;
}


/* Local functions */
int midiInAddBuffer(struct sblive_mididevice *midi_dev, struct midi_hdr **midihdrptr)
{
	struct midi_hdr *midihdr;

	midihdr = (struct midi_hdr *) kmalloc(sizeof(struct midi_hdr *), GFP_KERNEL);
	if (midihdr == NULL)
	  return -EINVAL;

	DPD(3, "kmalloc: [%p]\n", midihdr);
	
	initMidiHdr(midihdr);
	
	midihdr->lpData = (u8 *) kmalloc(MIDIIN_BUFLEN, GFP_KERNEL);	
	if (midihdr->lpData == NULL) 
	{
		DPF(2, "midihdr->lpData alloc fail.\n");

		kfree(midihdr);
		DPD(3, "kfree: [%p]\n", midihdr);

		return CTSTATUS_ERROR;
	}

	DPD(3, "kmalloc: [%p]\n", midihdr->lpData);
	
	if (sblive_mpuinAddBuffer(midi_dev->sb_hw->card_mpuin, midihdr) != CTSTATUS_SUCCESS) 
	{
		DPF(2, "sblive_mpuinAddBuffer failed.\n");

		kfree(midihdr->lpData);
		DPD(3, "kfree: [%p]\n", midihdr->lpData);

		kfree(midihdr);
		DPD(3, "kfree: [%p]\n", midihdr);

		return CTSTATUS_ERROR;
	}
	
	*midihdrptr = midihdr;
	osListAttach((struct sblive_list **) &midi_dev->pmiHdrList, (struct sblive_list *) midihdr);

	return CTSTATUS_SUCCESS;
}

void initMidiHdr(struct midi_hdr * midihdr)
{
	midihdr->bufferlength = MIDIIN_BUFLEN;
	midihdr->bytesrecorded = 0;
	midihdr->flags = 0;
	midihdr->next = NULL;
}

/* End of local functions */


int midi_init(struct sblive_hw *sb_hw)
{
	/* Initialize CardMpuOut struct */
	sb_hw->card_mpuout = kmalloc(sizeof(struct sblive_mpuout), GFP_KERNEL);
	if (sb_hw->card_mpuout == NULL) 
	{
		printk(KERN_WARNING "alloc struct sblive_mpuout: out of memory\n");
		return CTSTATUS_ERROR;
	}
	
	DPD(3, "kmalloc: [%p]\n", sb_hw->card_mpuout);
	
	sblive_mpuoutInit(sb_hw->card_mpuout, sb_hw);

	/* Initialize CardMpuIn struct */
	sb_hw->card_mpuin = kmalloc(sizeof(struct sblive_mpuin), GFP_KERNEL);
	if (sb_hw->card_mpuin == NULL) 
	{
		printk(KERN_WARNING "alloc struct sblive_mpuin: out of memory\n");
		return CTSTATUS_ERROR;
	}
	
	DPD(3, "kmalloc: [%p]\n", sb_hw->card_mpuin);
	
	sblive_mpuinInit(sb_hw->card_mpuin, sb_hw);

	return CTSTATUS_SUCCESS;
}

int midi_exit(struct sblive_hw *sb_hw)
{
	sblive_mpuoutExit(sb_hw);
	sblive_mpuinExit(sb_hw);
	kfree(sb_hw->card_mpuout);
	kfree(sb_hw->card_mpuin);
	
	DPD(3, "kfree: [%p]\n", sb_hw->card_mpuout);
	DPD(3, "kfree: [%p]\n", sb_hw->card_mpuin);
	
	sb_hw->card_mpuout = NULL;
	sb_hw->card_mpuin = NULL;

	return CTSTATUS_SUCCESS;
}


/* MIDI file operations */
struct file_operations emu10k1_midi_fops = 
{
	NULL,
	&emu10k1_midi_read,
	&emu10k1_midi_write,
	NULL,			/* readdir */
	&emu10k1_midi_poll,
	NULL,			/* ioctl */
	NULL,			/* mmap */
	&emu10k1_midi_open,
	NULL,			/* flush */
	&emu10k1_midi_release,
	NULL,			/* fsync */
	NULL,			/* fasync */
};
