/*
 **********************************************************************
 *     osutils.c - OS Services layer for emu10k1 driver
 *     Copyright 1999, 2000 Creative Labs, Inc.
 *
 **********************************************************************
 *
 *     Date                 Author          Summary of changes
 *     ----                 ------          ------------------
 *     October 20, 1999     Bertrand Lee    base code release
 *     November 2, 1999     Alan Cox        cleaned up
 *
 **********************************************************************
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 *     USA.
 *
 **********************************************************************
 */

/* FIXME! This file should not exist!
 * We do not need the crossplatform support anymore
 * This is a Linux-only driver
 */

#include "hwaccess.h"
#include "mycommon.h"


int osAllocMemPhysical(u32 size, struct memhandle **handle)
{
	u32 reqpage, order;

	*handle = (struct memhandle *) kmalloc(sizeof(struct memhandle), GFP_KERNEL);
	if (*handle == NULL)
		return CTSTATUS_NOMEMORY;

	DPD(3, "kmalloc: [%p]\n", *handle);

	order = 0;
	reqpage = size / PAGE_SIZE;

	if (size % PAGE_SIZE)
		reqpage++;

	if (reqpage != 0)
	{
		reqpage--;
		while (reqpage > 0)
		{
			reqpage >>= 1;
			order++;
		}
	}

	if (((*handle)->virtaddx = (void *)__get_free_pages(GFP_KERNEL, order)) == 0)
	{
		kfree((void *) *handle);

		DPD(3, "kfree: [%p]\n", *handle);
		*handle = NULL;
		return CTSTATUS_NOMEMORY;
	}

	DPD(3, "__get_free_pages: [%p]\n", (*handle)->virtaddx);

	/* in linux, we can directly access physical address, don't need to do
	 * phys_to_virt.
	 * In linux kernel 2.0.36, virt_to_bus does nothing, get_free_pages
	 * returns physical address. But in kernel 2.2.1 upwards,
	 * get_free_pages returns virtual address, we need to convert it
	 * to physical address. Then this physical address can be used to
	 * program hardware registers. */
	(*handle)->physaddx = virt_to_bus((*handle)->virtaddx);
	(*handle)->order    = order;

	return CTSTATUS_SUCCESS;
}

int osFreeMemPhysical(struct memhandle **handle)
{
	if (*handle == NULL)
		return CTSTATUS_ERROR;

	free_pages((unsigned long) (*handle)->virtaddx, (*handle)->order);
	kfree((void*) *handle);

	DPD(3, "free_pages: [%p]\n", (*handle)->virtaddx);
	DPD(3, "kfree: [%p]\n", *handle);

	*handle = NULL;
	return CTSTATUS_SUCCESS;
}


int osListAttach(struct sblive_list **head, struct sblive_list *n)
{
	unsigned long flags;

	if (head == NULL)
		return CTSTATUS_ERROR;

	spin_lock_irqsave(&sblive_spinlock, flags);

	n->next = NULL;

	if (*head != NULL)
	{
		struct sblive_list *t = *head;

		while (t->next != NULL)
			t = t->next;
		t->next = n;
	} else
		*head = n;

	spin_unlock_irqrestore(&sblive_spinlock, flags);

	return CTSTATUS_SUCCESS;
}


int osListRemove(struct sblive_list **head, struct sblive_list *dead)
{
	u32 status = CTSTATUS_SUCCESS;
	unsigned long flags;

	if (head == NULL)
		return CTSTATUS_ERROR;

	spin_lock_irqsave(&sblive_spinlock, flags);

	if (*head != dead)
	{
		struct sblive_list *t = *head;

		while (t != NULL && t->next != dead)
			t = t->next;

		if (t != NULL)
		{
			t->next = dead->next;
			dead->next = NULL;
		} else
			status = CTSTATUS_ERROR;
	} else
		*head = dead->next;

	spin_unlock_irqrestore(&sblive_spinlock, flags);

	return status;
}


struct sblive_list *osListGetNext(struct sblive_list *head, struct sblive_list *curr)
{
	if (curr == NULL)
		return head;

	return curr->next;
}

